#!/usr/bin/python
#-----------------------------------------------------------------------
# File    : events.py
# Contents: collect events underlying found patterns
# Author  : Christian Borgelt
# History : 2014.07.01 file created
#           2014.07.03 png image creation added
#           2014.07.10 choice of colors, Metapost output added
#           2014.07.11 color decoding moved before image creation
#           2014.07.18 processing of sequences/ordered item sets added
#           2014.07.18 split into one figure per pattern added
#-----------------------------------------------------------------------
from sys     import argv, stderr, version_info
from time    import time
from math    import floor, ceil
from PIL     import Image
from coconad import getrec

#-----------------------------------------------------------------------
# Constants
#-----------------------------------------------------------------------
oo      = float('inf')          # positive infinity as a constant
heapmin = 7                     # minimum number of trains for a heap

#-----------------------------------------------------------------------
# Event Collection Functions
#-----------------------------------------------------------------------

def sift (heap, lft, rgt):
    '''sift (heap, lft, rgt)
Let an element sift down in a train heap.
heap    list of pairs (list of points/times, position)
        (a heap element may have additional entries, though)
lft     left  border of the heap (index of sift element)
rgt     right border of the heap (index of last element)'''
    h = heap[lft]               # note the sift element
    t,k = h[:2]; s = t[k]       # and get its current time
    i = lft +lft +1             # compute index of first successor
    while i <= rgt:             # sift loop
        t,k = heap[i][:2]       # get the first successor
        x = t[k]                # and from it the current point
        if i < rgt:             # if there is a second successor
            t,k = heap[i+1][:2] # and it is smaller than the first
            y = t[k]            # go to the second successor
            if y < x: x = y; i += 1
        if x >= s:              # if the successor is no less
            break               # than the sift element, abort
        heap[lft] = heap[i]     # let the successor ascend in heap
        lft = i                 # note index of current element and
        i  += i+1               # compute index of first successor
    heap[lft] = h               # store the sift element

#-----------------------------------------------------------------------

def events (trains, width, flat=1):
    '''events (trains, width, flat=1)
Collect synchronous event sets of a list of trains.
trains  list of pairs (item id, list of points/times)
width   width of time window/maximal distance (width >= 0)
flat    whether to collect events as a flat list (default: 1)
returns the coincident events of the item set underlying the trains'''
    k = len(trains)             # get the number of trains
    if k <= 1:                  # if there is only one train
        return [(i,s) for i,t in trains for s in t]
    events = []                 # initialize the event list
    if k <= 2:                  # if there are only two trains
        a,s = trains[0]         # get the two items
        b,t = trains[1]         # and the corresponding trains
        n,m = len(s),len(t)     # get the train lengths
        i = k = 0               # traverse the two trains
        while i < n and k < m:  # while not at end of either train
            if abs(s[i]-t[k]) <= width:
                if flat: events.extend([(a,s[i]),(b,t[k])])
                else:    events.append(((a,s[i]),(b,t[k])))
                i += 1; k += 1  # collect synchronous points/times
            elif s[i] < t[k]: i += 1
            else:             k += 1
        return events           # return the computed support
    heap = [[t,0,r] for r,t in trains if len(t) > 0]
    n    = len(heap)            # build a heap/array of all trains
    if n < k: return events     # and check for no empty trains
    if n >= heapmin:            # if there are many trains
        for i in reversed(range(n >> 1)):
            sift(heap, i, n-1)  # build a heap w.r.t. the first times
    tmax = max(t[0] for t,i,k in heap)
    while 1:                    # support computation loop
        t,i = heap[0][:2]; tmin = t[i]; k = 0
        if n < heapmin:         # if there are only few trains
            for j in range(1,n):  # find the train
                t,i = heap[j][:2] # with the smallest time
                if t[i] < tmin: tmin = t[i]; k = j
        if tmax-tmin > width:   # if the times are not synchronous,
            heap[k][1] += 1     # skip the processed time and
            t,i = heap[k][:2]   # check whether it was the last
            if i >= len(t): return events
            if t[i] > tmax: tmax = t[i] # update the maximum time
            if n >= heapmin: sift(heap, 0, n-1)
            continue            # let the new time sift down in heap
        x = tuple([(r,t[i]) for t,i,r in heap])
        if flat: events.extend(x) # collect the events
        else:    events.append(x) # of a synchronous group
        for h in heap:          # traverse all times in the array
            h[1] += 1           # skip the processed time and
            t,i = h[:2]         # check whether it was the last
            if i >= len(t): return events
            if t[i] > tmax: tmax = t[i] # update the maximum time
        if n >= heapmin:        # if needed, rebuild the heap
            for i in reversed(range(n >> 1)):
                sift(heap, i, n-1)

#-----------------------------------------------------------------------

def seqevents (trains, width, flat=1):
    '''seqevents (trains, width, flat=1)
Collect (synchronous) event sequences of a list of trains.
trains  list of pairs (item id, list of points/times)
width   width of time window/maximal distance (width >= 0)
flat    whether to collect events as a flat list (default: 1)
returns the coincident events of the item set underlying the trains'''
    k = len(trains)             # get the number of trains
    if k <= 1:                  # if there is only one train
        return [(i,s) for i,t in trains for s in t]
    events = []                 # initialize the event list
    if k <= 2:                  # if there are only two trains
        a,s = trains[0]         # get the two items
        b,t = trains[1]         # and the corresponding trains
        n,m = len(s),len(t)     # get the train lengths
        i = k = 0               # traverse the two trains and
        while i < n and k < m:  # while not at end of either train
            d = t[k]-s[i]       # check for synchronous points/times
            if   d < 0:     k += 1
            elif d > width: i += 1
            else:               # collect the synchronous sequence
                if flat: events.extend([(a,s[i]),(b,t[k])])
                else:    events.append(((a,s[i]),(b,t[k])))
                i += 1; k += 1  # advance the point positions
        return events           # return the collected events
    curr = [[t,0,r] for r,t in trains if len(t) > 0]
    n    = len(curr)            # build a position array of all trains
    if n < k: return events     # and check for no empty trains
    a,z = curr[0],curr[-1]      # get first and last train/position
    while a[1] < len(a[0]):     # while not at end of first train
        for i in range(1,n):    # traverse the following trains
            b,c = curr[i-1:i+1] # (or rather consecutive pairs)
            while c[1] < len(c[0]) and c[0][c[1]] < b[0][b[1]]:
                c[1] += 1       # ensure item succession in time
            if c[1] >= len(c[0]):
                return events   # if at the end of a train, abort
        if z[0][z[1]] -a[0][a[1]] <= width:
            x = tuple([(r,t[i]) for t,i,r in curr])
            if flat: events.extend(x) # collect the events
            else:    events.append(x) # of a synchronous group
        a[1] += 1               # get the next starting point
    return events               # return the collected events

#-----------------------------------------------------------------------
# Image Functions
#-----------------------------------------------------------------------

def prepare (trains, pats, seq=False, rat4p=False):
    k = 1                       # check whether all trains
    for i in trains:            # have items/identifiers
        try: int(i)             # that are integer numbers
        except ValueError: k = 0; break
    if k: items = sorted([i for i in trains], key=lambda x: int(x))
    else: items = sorted([i for i in trains])
    if rat4p:                   # if to rearrange trains for patterns
        iips = set([i for p,c in pats for i in p])
        rest = [i for i in items if i not in iips]
        xpts = [(frozenset(p),c*len(p)) for p,c in pats]
        iseq = []               # turn item sets into actual sets
        while 1:                # while to remove more items
            xpts = [p for p in xpts if p[0] <= iips]
            if not xpts: break  # reduce patterns, init. evaluations
            vals = dict([(i,0) for i in iips])
            for p,c in xpts:    # sum item instance counts per item
                for i in p: vals[i] += c
            n = min(vals, key=lambda x: vals[x])
            c = vals[n]         # find the least connected item and
            iseq.append(n)      # add the item to the sequence, then
            iips.remove(n)      # remove it from the remaining items
        iseq += iips            # add the last items
        iseq.reverse()          # reverse the created sequence
        items = iseq +rest      # and add the remaining items
    evcfn = seqevents if seq else events
    items = dict([(items[k],k) for k in range(len(items))])
    evall = set ([(items[i],s) for i in trains for s in trains[i]])
    evpat = set ([(items[i],s) for p,c in pats
                  for i,s in evcfn([(i,trains[i]) for i in p], width)])
    return (evall,evpat)        # collect events to display

#-----------------------------------------------------------------------

def pngimg (fname, trains, pats, seq, joint, beg, end, sx, sy,
            evtcol=(0.67,0.67,0.67), patcol=(0.8,0,0),
            bkgnd=(1.0,1.0,1.0), rat4p=False):
    p = [s for i in trains for s in trains[i]]
    if beg <= -oo: beg = floor(min(p))
    if end >= +oo: end = ceil (max(p))
    bkgnd  = tuple([int(floor(255*x+0.5)) for x in bkgnd])
    evtcol = tuple([int(floor(255*x+0.5)) for x in evtcol])
    patcol = tuple([int(floor(255*x+0.5)) for x in patcol])
    h      = len(trains)*sy     # get height and width of image
    w      = int(ceil((end-beg)*sx))
    for p in [pats] if joint else [[p] for p in pats]:
        img = Image.new('RGB', (w,h), bkgnd)
        pxl = img.load()        # create the output image
        evall,evpat = prepare(trains, p, seq, rat4p)
        for i,s in evall-evpat: # traverse all events/spikes
            x = int((s-beg)*sx) # draw a vertical bar (grey by default)
            y = h-(i+1)*sy
            for k in range(int(sy)): pxl[x,y+k] = evtcol
        for i,s in evpat:       # traverse pattern events/spikes
            x = int((s-beg)*sx) # draw a vertical bar (red by default)
            y = h-(i+1)*sy
            for k in range(int(sy)): pxl[x,y+k] = patcol
        if len(p) <= 1: n = '-'.join(p[0][0])
        else: n = ''            # form name of pattern
        with open(fname+n+'.png', 'wb') as out:
            img.save(out,'png') # save the created image

#-----------------------------------------------------------------------

def mpsrc (fname, trains, pats, seq, joint, beg, end, sx, sy, thick,
           evtcol=(0.667,0.667,0.667), patcol=(0.8,0,0),
           bkgnd=(1.0,1.0,1.0), rat4p=False,
           xlabel='"time"', ylabel='"neurons"'):
    p = [s for i in trains for s in trains[i]]
    if beg <= -oo: beg = floor(min(p))
    if end >= +oo: end = ceil (max(p))
    bkgnd = tuple(bkgnd)        # build a list of pattern (sets)
    white = (bkgnd[0] == 1) and (bkgnd[1] == 1) and (bkgnd[2] == 1)
    with open(fname+'.mp', 'w') as out:
        out.write('prologues := 2;\n\n')
        out.write('defaultfont := "ptmr8r";\n')
        out.write('labeloffset := 0pt;\n\n')
        out.write('numeric sx; sx := %.6fpt;\n' % sx)
        out.write('numeric sy; sy := %.6fpt;\n' % sy)
        out.write('numeric pw; pw := %.6fpt;\n' % thick)
        out.write('color e; e := (%.6f,%.6f,%.6f);\n' % tuple(evtcol))
        out.write('color p; p := (%.6f,%.6f,%.6f);\n' % tuple(patcol))
        out.write('path fp;\n')
        out.write('\n')
        out.write('vardef bar (expr x, y, c) =\n')
        out.write('  draw (x*sx,y*sy+.75pw)'
                      +'--(x*sx,(y+1)*sy-.75pw) withcolor c;\n')
        out.write('enddef;\n')
        out.write('\n')         # print the file header
        fid = 0                 # initialize the figure identifier
        for p in [pats] if joint else [[p] for p in pats]:
            if len(p) <= 1: n = '-'.join(p[0][0])
            else: n = ''        # traverse the patterns, form name
            evall,evpat = prepare(trains, p, seq, rat4p)
            out.write('beginfig(%d); %% %s\n' % (fid,n))
            h = len(trains)*sy  # get height and width of image
            w = (end-beg)*sx    # and draw the background
            out.write('  fp := unitsquare xscaled (%.6f+pt)' % w
                                       +' yscaled %.6f\n'    % h
                     +'        shifted (-0.5pt,-0.25pt);\n')
            if not white:       # if non-white background
                out.write('  fill fp withcolor '
                         +'(%.6f,%.6f,%.6f);\n' % bkgnd)
            out.write('  pickup pencircle scaled 0.5pt; draw fp;\n')
            if xlabel or ylabel:
                if xlabel:
                    out.write('  label.lrt(%s,' % xlabel
                             +' (0,-0.7mm));\n')
                if ylabel:
                    out.write('  label.urt(%s,' % ylabel
                             +' (0, 0.8mm)) rotated 90;\n')
            out.write('  pickup pensquare scaled pw;\n');
            for i,s in evall-evpat: # traverse all events/spikes
                out.write('  bar(%.6f,%.6f,e);\n' % (s-beg,i))
            for i,s in evpat:   # traverse pattern events/spikes
                out.write('  bar(%.6f,%.6f,p);\n' % (s-beg,i))
            out.write('endfig;\n\n')
            fid += 1            # terminate the figure
        out.write('end.\n')     # terminate the Metapost file

#-----------------------------------------------------------------------
# Main Program
#-----------------------------------------------------------------------

def error (msg):
    stderr.write(msg); exit()   # print an error message and abort

#-----------------------------------------------------------------------

if __name__ == '__main__':
    desc    = 'collect events underlying given patterns'
    version = 'version 1.3 (2014.07.18)         ' \
            + '(c) 2014   Christian Borgelt'
    args    = []                # list of program arguments/options
    opts    = {'w': ['width', 0.003],  # width of window/region
               'o': ['seq',   False],  # flag for item sequences
               'a': ['beg',  -oo   ],  # minimum of point/time span
               'z': ['end',   oo   ],  # minimum of point/time span
               'i': ['img',   ''   ],  # name of png image file
               'm': ['mpost', ''   ],  # name of Metapost  file
               'j': ['joint', False],  # joint image with all patterns
               's': ['scale', 300.0],  # scale for horizontal axis
               'h': ['height',3    ],  # height of an image row (pixels)
               'e': ['evtcol', '0.67:0.67:0.67'],
               'p': ['patcol', '0.8:0:0'],   # event/spike colors
               'g': ['bkgnd',  '1:1:1'],     # background color
               't': ['thick', 0.5  ],  # pen thickness
               'x': ['rat4p', False],  # rearrange trains for patterns
               'l': ['train', False],  # one train per line/record
               'y': ['item',  True ],  # item in line, item first
               'r': ['recseps', '\\n' ],     # record  separators
               'f': ['fldseps', ' ,\\t' ],   # field   separators
               'b': ['blanks',  ' \\t\\r' ], # blank   characters
               'C': ['comment', '#' ] }      # comment characters

    if len(argv) <= 1:          # if no arguments are given
        opts = dict([(o,opts[o][1]) for o in opts])
        print('usage: events.py [options] trains pats events')
        print(desc); print(version)
        print('-w#      width of time window/maximum distance    '
                      +'(default: %g)' % opts['w'])
        print('-o       find item sequences (respect item order) '
                      +'(default: sets)');
        print('-a#      start of point/time range                '
                      +'(default: auto)')
        print('-z#      end   of point/time range                '
                      +'(default: auto)')
        print('-i#      name of png image(s) to create           '
                      +'(default: none)')
        print('-m#      name of Metapost file to create          '
                      +'(default: none)')
        print('-j       one joint image for all patterns         '
                      +'(default: separate)')
        print('-s#      (time) scale for horizontal image axis   '
                      +'(default: %g)' % opts['s'])
        print('-h#      height of an image row in pixels         '
                      +'(default: %d)' % opts['h'])
        print('-e#:#:#  color of background events/spikes        '
                      +'(default: %s)' % opts['e'])
        print('-p#:#:#  color of pattern    events/spikes        '
                      +'(default: %s)' % opts['p'])
        print('-g#:#:#  background color (all colors as RGB)     '
                      +'(default: %s)' % opts['g'])
        print('-t#      pen thickness for Metapost (in pt)       '
                      +'(default: %g)' % opts['t'])
        print('-x       rearrange trains w.r.t. patterns         '
                      +'(default: %s)' % str(opts['x']))
        print('-l       one train per line (item + points)       '
                      +'(default: id/time pairs)');
        print('-y       no item (with -l) or item after point    '
                      +'(default: item first)');
        print('-r#      record  separators                       '
                      +'(default: "%s")' % opts['r'])
        print('-f#      field   separators                       '
                      +'(default: "%s")' % opts['f'])
        print('-b#      blank   characters                       '
                      +'(default: "%s")' % opts['b'])
        print('-C#      comment characters                       '
                      +'(default: "%s")' % opts['C'])
        print('trains   file to read trains from                 '
                      +'[required]')
        print('pats     file to read patterns from               '
                      +'[required]')
        print('events   file to write collected events to        '
                      +'[required]')
        exit()                  # print usage message and abort

    stderr.write('events.py')   # print a startup message
    stderr.write(' - ' +desc +'\n' +version +'\n')

    # --- evaluate program options and arguments ---
    for a in argv[1:]:          # traverse the program arguments
        if a[0] != '-':         # check for a fixed argument
            if len(args) < 3: args.append(a); continue
            else: error('too many arguments\n')
        if a[1] not in opts:    # check for an option
            error('unknown option: %s\n' % a[:2])
        o = opts[a[1]]          # get the corresponding option
        v = a[2:]               # and get the option argument
        if   isinstance(o[1],True.__class__): o[1] = not o[1]
        elif isinstance(o[1], (0).__class__): o[1] = int(v)
        elif isinstance(o[1], 0.0.__class__): o[1] = float(v)
        else:                                 o[1] = v
    if len(args) < 3: error('missing arguments\n')
    opts = dict([opts[o] for o in opts])

    width   = opts['width']     # time window width/maximum distance
    seq     = opts['seq']       # flag for item sequences
    joint   = opts['joint']     # flag for a joint image
    beg     = opts['beg']       # beginning of allowed span of spikes
    end     = opts['end']       # end       of allowed span of spikes
    tpr     = opts['train']     # flag for one train per line
    item    = opts['item']      # and flag for item first
    recseps = opts['recseps']   # record  separators
    fldseps = opts['fldseps']   # field   separators
    blanks  = opts['blanks']    # blank   characters
    comment = opts['comment']   # comment characters
    x = [recseps,fldseps,blanks,comment]
    if version_info[0] >= 3:    # decode ASCII escape sequences
        x = [bytes(s, 'utf-8').decode('unicode_escape') for s in x]
    else:                       # decode ASCII escape sequences
        x = [s.decode('string_escape') for  s in x]
    recseps,fldseps,blanks,comment = x

    # --- read the train set ---
    t = time()                  # start timer, print log message
    stderr.write('reading %s ... ' % args[0])
    trains = dict(); k = 1      # initialize the train dictionary
    with open(args[0], 'rb') as inp:
        while 1:                # record read loop
            rec = getrec(inp, recseps, fldseps, blanks, comment)
            if rec == None: break
            if not rec: continue# read the next record
            if tpr:             # if one train per line/record
                if item: n = rec[0]; rec = rec[1:]
                else:    n = k+1# get the item/neuron identifier
                trains[n] = [x for x in [float(x) for x in line]
                             if x >= beg and x <= end]
                k += 1          # collect the trains
            else:               # if item/ point pairs
                if item: n,s = rec
                else:    s,n = rec
                s = float(s)    # collect points per item/neuron
                if s < beg or s > end: continue
                if n in trains: trains[n].append(s)
                else:           trains[n] = [s]
    points = sum([len(trains[n]) for n in trains])
    stderr.write('[%d item(s), %d point(s)]' % (len(trains), points))
    stderr.write(' done [%.2fs].\n' % (time()-t))

    # --- read the pattern set ---
    t = time()                  # start timer, print log message
    stderr.write('reading %s ... ' % args[1])
    pats = []                   # initialize the list of patterns
    with open(args[1], 'rb') as inp:
        while 1:                # record read loop
            rec = getrec(inp, recseps, fldseps, blanks, comment)
            if not rec: break   # read the next record
            pats.append((rec[:-1],int(rec[-1][1:-1])))
    stderr.write('[%d pattern(s)]' % len(pats))
    stderr.write(' done [%.2fs].\n' % (time()-t))

    # --- collect events ---
    t = time()                  # start timer, print log message
    stderr.write('writing %s ... ' % args[2])
    with open(args[2], 'w') as out:
        k = 0                   # initialize the event counter
        for p,c in pats:        # traverse the patterns
            e = seqevents([(n,trains[n]) for n in p], width) if seq \
              else events([(n,trains[n]) for n in p], width)
            for x in e[:-1]: out.write('%s:%.15g ' % x)
            if len(e) > 0: out.write('%s:%.15g\n' % e[-1])
            k += len(e)         # collect and write the events
    stderr.write('[%d pattern(s), %d point(s)]' % (len(pats),k))
    stderr.write(' done [%.2fs].\n' % (time()-t))

    # --- create and write png image ---    
    bkgnd  = tuple([float(x) for x in opts['bkgnd'].split(':')])
    evtcol = tuple([float(x) for x in opts['evtcol'].split(':')])
    patcol = tuple([float(x) for x in opts['patcol'].split(':')])
    fname  = opts['img']        # get name of png image file
    if fname[-4:] == '.png': fname = fname[:-4]
    if len(fname) > 0:          # if a file name is given
        t = time()              # start timer, print log message
        stderr.write('writing %s ... ' % fname)
        pngimg(fname, trains, pats, seq, joint, beg, end,
               opts['scale'], opts['height'],
               evtcol, patcol, bkgnd, opts['rat4p'])
        n = sum([len(trains[i]) for i in trains])
        stderr.write('[%d/%d point(s)]' % (n,k))
        stderr.write(' done [%.2fs].\n' % (time()-t))

    # --- create and write Metapost source ---    
    fname = opts['mpost']       # get name of Metapost file
    if fname[-3:] == '.mp': fname = fname[:-3]
    if len(fname) > 0:          # if a file name is given
        t = time()              # start timer, print log message
        stderr.write('writing %s.mp ... ' % fname)
        mpsrc(fname, trains, pats, seq, joint, beg, end,
              opts['scale'], opts['height'], opts['thick'],
              evtcol, patcol, bkgnd, opts['rat4p'])
        n = sum([len(trains[i]) for i in trains])
        stderr.write('[%d/%d point(s)]' % (n,k))
        stderr.write(' done [%.2fs].\n' % (time()-t))
