#!/usr/bin/python
#-----------------------------------------------------------------------
# File    : diags.py
# Contents: create diagrams for hypernad experiments
# Author  : Christian Borgelt
# History : 2014.07.10 file created
#-----------------------------------------------------------------------
from sys     import argv, path
from os.path import join
from math    import log10
UTILDIR  = join('..', 'util')
if UTILDIR  not in path: path.insert(0, UTILDIR)
from id2par  import get_args

tics   = [ 0,  1,  2,  3,  4,  5,  6,  7,  8,  9,
          10, 11, 12, 13, 14, 15, 16, 17, 18, 19,
          20, 21, 22, 23, 24, 25, 26, 27, 28, 29 ]

titles = { 'fn': 'false neg.',
           'fx': 'false neg.',
           'sp': 'supersets',
           'sb': 'subsets',
           'ov': 'overlap',
           'ur': 'unrelated',
           'mt': 'empty' }

#-----------------------------------------------------------------------

def header (out, pro, addinit=''):
    out.write('verbatimtex\n')
    out.write('\\documentclass[12pt]{article}\n')
    out.write('\\usepackage{mathptmx}\n')
    out.write('\\def\\hnorm{\\hbox to0pt{\\phantom{g)}\\hss}}\n')
    out.write('\\begin{document}\n')
    out.write('etex\n\n')
    out.write('prologues := ' +str(pro) +';\n\n')
    out.write('numeric expand; expand = 1/3;\n\n')
    out.write('input axes;\n')
    out.write('input barchart;\n\n')
    out.write('vardef init_chart (expr xoff, yoff, zoff) =\n')
    out.write('  numeric data[][], n; n = 0;\n')
    out.write('  barchart.set_defaults("ptmr8r");\n')
    out.write('  barchart.set_view (-120, 36);\n')
    out.write('  barchart.set_light(-120, 75);\n')
    out.write('  barchart.clipcolor    := 0.7white;\n')
    out.write('  axes.title.size       := 1.2;  % 12pt\n')
    out.write('  axes.title.skip       := 1.1*fontsize "ptmr8r";\n')
    out.write('  axes.x.tlabel.size    := 8/10; %  8pt\n')
    out.write('  axes.x.alabel.size    := 1;    % 10pt\n')
    out.write('  axes.x.alabel.offset  := xoff;\n')
    out.write('  axes.y.tlabel.size    := 8/10; %  8pt\n')
    out.write('  axes.y.alabel.size    := 1;    % 10pt\n')
    out.write('  axes.y.alabel.offset  := yoff;\n')
    out.write('  axes.z.tlabel.size    := 8/10; %  8pt\n')
    out.write('  axes.z.alabel.size    := 1;    % 10pt\n')
    out.write('  axes.z.alabel.offset  := zoff;\n')
    out.write('  axes.z.grid.thickness := 0.4pt;\n')
    out.write('  axes.z.kinked         := true;\n')
    out.write(addinit)
    out.write('enddef;\n\n')
    out.write('vardef set_data (text list) =\n')
    out.write('  barchart.set_data(data[n], list);\n')
    out.write('  n := n+1;\n')
    out.write('enddef;\n\n')

#-----------------------------------------------------------------------

def footer (out):
    out.write('end.\n')

#-----------------------------------------------------------------------

def readfile (fname):
    with open(fname, 'r') as inp:
        c = [l.split() for l in inp]
    return dict([((int(x),int(y)),float(z)) for x,y,z in c])

#-----------------------------------------------------------------------

def diag (out, fid, pfx='fx',
          ncnt=100, period=3, rate=20, msel=0, jitter=0.003,
          width=0.003, logth=0, mode=('p',0)):
    name = pfx+'_n%d_p%g_r%g_m%d_j%g_w%g_t%g_%c%d' \
         % (ncnt,period,rate,msel,jitter*1000,width*1000,logth,
            mode[0],mode[1])
    data = readfile('data/'+name+'.txt')
    out.write('beginfig(' +str(fid) +'); % '+name+'\n')
    out.write('  init_chart((0,4.0mm), (0,4.4mm), (0,5.2mm));\n')
    out.write('  axes.title.offset    := (0,1mm);\n')
    out.write('  barchart.null.max    := -0.5;\n')
    out.write('  barchart.invalid.max := -1.5;\n')
    out.write('  barchart.invalid.barcolor := 0.5white;\n')
    out.write('  barchart.set_units(4mm, 4mm, 24mm);\n')
    out.write('  barchart.set_ranges((11,0.84,0.08), '
                                   +'(11,0.84,0.08), (0,1));\n')
    out.write('  barchart.draw_xlabel(0, '
               +'btex \\small pattern size~$z$\\hnorm etex);\n')
    out.write('  barchart.draw_xticks("'+str(tics[2])+'"')
    for x in tics[3:13]: out.write(', "'+str(x)+'"')
    out.write(');\n')
    out.write('  barchart.draw_ylabel(0, '
               +'btex \\small instances~$c$\\hnorm etex);\n')
    r = range(2,13) if msel == 0 else range(2*msel+1,2*msel+12)
    out.write('  barchart.draw_yticks("'+str(tics[r[0]])+'"')
    for y in r[1:]: out.write(', "'+str(tics[y])+'"')
    out.write(');\n')
    out.write('  barchart.draw_zlabel(0, '
               +'btex rate\\hnorm etex);\n')
    out.write('  barchart.draw_zticks(0, 0.2, 0.4, 0.6, 0.8, 1);\n')
    name = '%d $10^%g$ %c %d' % (msel,logth,mode[0],mode[1])
    out.write('  barchart.draw_title(-3, -1, '
               +'btex '+titles[pfx]+'\\hnorm etex, '
               +'btex '+name+'\\hnorm etex);\n')
    out.write('  barchart.draw_frame;\n')
    for x in range(2,13):        # traverse the x-values (sizes)
        out.write('  set_data(') # start the next data row
        first = True             # set flag for first value
        for y in r:              # traverse the y-values (instances)
            if not first: out.write(', ')
            else:         first = False
            z = data[(x,y)] if (x,y) in data else 0
            out.write('%f' % (-1 if z <= 0 else z if z <= 1 else 2))
        out.write(');\n')       # print the values/fill data matrix
    out.write('  barchart.draw_data(data, 0, 5, expand);\n')
    out.write('endfig;\n\n')    # draw the diagram

#-----------------------------------------------------------------------

if __name__ == '__main__':
    grids = [ [ ( 'pfx',    ['fn','fx','sp','sb','ov','ur','mt'] ),
                ( 'ncnt',   [   100 ] ),
                ( 'period', [     3 ] ),
                ( 'rate',   [    20 ] ),
                ( 'jitter', [ 0.003 ] ),
                ( 'width',  [ 0.003 ] ),
                ( 'logth',  [ 0, 1, 2, 3 ] ),
                ( 'msel',   [ 0, 1, 2, 3, 4, 5 ] ),
                ( 'mode',   [('p',0),('p',1),
                             ('i',0),('i',1),
                             ('s',0),('s',1),] ) ] ]
    fid = 0
    with open(argv[1] if len(argv) > 1 else 'hypernad.mp', 'w') as out:
        header(out, 2)
        for i in range(get_args(grids)):
            diag(out, fid, **get_args(grids, i)); fid += 1
        footer(out)
